#include <memory.h>
#include <stdio.h>
#include "interval.h"

/* need to do own pseudo-stack for mem allocation */
static Set arena=NULL, next=NULL;
unsigned arena_size = 0;

void
reset_sets()
{
	next = arena;
}

/* create an interval at the head of a set */
Set
_MakeSet(s,e,n)
float s,e;
Set n;
{
        if (next==arena+arena_size) {
                if (arena_size) {
                        register Set p;
                        int delta;
                        arena_size *= 2;
                        p = (Set)realloc(arena,arena_size*sizeof (struct Intrvl));
                        if (!p) {
                                fprintf(stdout,"Out of memory!\n");
                                exit(1);
                        }
                        delta = (int)p - (int)arena;
                        arena = p;
                        if (delta) {
                                (int)next += delta;
                                while (p<next) {
                                        if (p->next) (int)p->next += delta;
                                        p++;
                                }
                        }
                } else {
                        arena_size = 512/sizeof (struct Intrvl);
                        arena=next=(Set)malloc(arena_size*sizeof (struct Intrvl));
                }
        }
        {
                register Set i;
                i = next++;
                i->start.val = s;
                i->end.val = e;
                i->next = n;
                return i;
        }
}

/* find the union of two Sets */
Set
USet(a,b)
Set a,b;
{
        /* union of a set with itself returns the set */
        if (a==b) return a;
        /* union of a set with the empty set returns the set */
        if (a==NULL) return b;
        if (b==NULL) return a;
        /* check for union of disjoint sets */
        if (a->end.val < b->start.val) {
                return _MakeSet(a->start.val,a->end.val,USet(a->next,b));
        }
        if (b->end.val < a->start.val) {
                return _MakeSet(b->start.val,b->end.val,USet(a,b->next));
        }
        /* union of a set with a subset returns the set */
        if ((a->start.val <= b->start.val) && (a->end.val >= b->end.val)) {
                if (b->next==NULL) return a;
                return _MakeSet(a->start.val,a->end.val,USet(a->next,b->next));
        }
        if ((a->start.val >= b->start.val) && (a->end.val <= b->end.val)) {
                if (a->next==NULL) return b;
                return _MakeSet(b->start.val,b->end.val,USet(a->next,b->next));
        }
        /* overlapping sets return the combined interval */
		{
			float s = (a->start.val<b->start.val)?a->start.val:b->start.val;
			float e = (a->end.val>b->end.val)?a->end.val:b->end.val;
			a = a->next;
			b = b->next;
			while (1) {
				if (a && (a->start.val < e)) {
					e = a->end.val;
					a = a->next;
				} else if (b && (b->start.val < e)) {
					e = b->end.val;
					b = b->next;
				} else {
					break;
				}
			}
        	return _MakeSet(s, e, USet(a,b) );
		}
}

/* find the intersection of two Sets */
Set
ISet(a,b)
Set a,b;
{
        /* intersection of a set with itself returns the set */
        if (a==b) return a;
        /* intersection of a set with the empty set returns the empty set */
        if ((a==NULL) || (b==NULL)) return NULL;
        /* intersection of disjoint sets is the empty set */
        if (a->end.val < b->start.val) {
                return ISet(a->next,b);
        }
        if (b->end.val < a->start.val) {
                return ISet(a,b->next);
        }
        /* intersection of a set with a subset returns the subset */
        if ((a->start.val >= b->start.val) && (a->end.val <= b->end.val)) {
                return USet(a,ISet(a->next,b));
        }
        if ((a->start.val <= b->start.val) && (a->end.val >= b->end.val)) {
                return USet(b,ISet(a,b->next));
        }
        /* overlapping sets return the common interval */
        if (a->end.val > b->end.val) {
                return _MakeSet(a->start.val,b->end.val,ISet(a,b->next));
        } else {
                return _MakeSet(b->start.val,a->end.val,ISet(a->next,b));
        }
}

/* find the difference of two Sets */
Set
DSet(a,b)
Set a,b;
{
        /* difference of a set with itself returns the empty set */
        if (a==b) return NULL;
        /* difference of the empty set with a set returns the empty set */
        if (a==NULL) return NULL;
        /* difference of a set with the empty set returns the set */
        if (b==NULL) return a;
        /* difference of disjoint sets isn't too hard */
        if (a->end.val < b->start.val) {
                return USet(a,DSet(a->next,b));
        }
        if (b->end.val < a->start.val) {
                return DSet(a,b->next);
        }
        /* difference of a set with a subset returns the subset */
        if ((a->start.val >= b->start.val) && (a->end.val <= b->end.val)) {
                return DSet(a->next,b);
        }
        if ((a->start.val <= b->start.val) && (a->end.val >= b->end.val)) {
                return _MakeSet(a->start.val,b->start.val,DSet(
                                _MakeSet(b->end.val,a->end.val,a->next),b->next));
        }
        /* overlapping sets return the common interval */
        if (a->end.val > b->end.val) {
                return DSet(_MakeSet(b->end.val,a->end.val,a->next),b->next);
        } else {
                return _MakeSet(a->start.val,b->start.val,DSet(a->next,b));
        }
}

#ifdef TESTBED
/* print the contents of a set */
void
PrintSet(f,s,i)
FILE *f;
char *s;
Set i;
{
        fprintf(f,"Set %s is",s);
        if (!i) {
                fprintf(f," empty");
        } else {
                while (i) {
                        fprintf(f," [%g..%g]",i->start.val,i->end.val);
                        i = i->next;
                }
        }
        fprintf(f,".\n");
        return;
}

/* test the routines */
int
main()
{
        Set a,b,c,d,e,f,g;
        a = MakeSet(1.0,5.0);
        b = MakeSet(3.0,7.0);
        c = MakeSet(6.0,9.0);
        d = USet(a,c);
        e = NULL;
        f = _MakeSet(1.0,4.0,MakeSet(5.0,8.0));
        g = _MakeSet(0.0,2.0,_MakeSet(3.0,6.0,MakeSet(7.0,9.0)));
        PrintSet(stdout,"a",a);
        PrintSet(stdout,"b",b);
        PrintSet(stdout,"c",c);
        PrintSet(stdout,"d",d);
        PrintSet(stdout,"e",e);
        PrintSet(stdout,"f",f);
        PrintSet(stdout,"g",g);
        printf("memory used = %d\n",next-arena);
        PrintSet(stdout,"a U a",USet(a,a));
        PrintSet(stdout,"a U b",USet(a,b));
        PrintSet(stdout,"b U a",USet(b,a));
        PrintSet(stdout,"a U c",USet(a,c));
        PrintSet(stdout,"c U a",USet(c,a));
        PrintSet(stdout,"a U e",USet(a,e));
        PrintSet(stdout,"e U a",USet(e,a));
        PrintSet(stdout,"a ^ a",ISet(a,a));
        PrintSet(stdout,"a ^ b",ISet(a,b));
        PrintSet(stdout,"b ^ a",ISet(b,a));
        PrintSet(stdout,"a ^ c",ISet(a,c));
        PrintSet(stdout,"c ^ a",ISet(c,a));
        PrintSet(stdout,"a ^ e",ISet(a,e));
        PrintSet(stdout,"e ^ a",ISet(e,a));
        PrintSet(stdout,"a - a",DSet(a,a));
        PrintSet(stdout,"a - b",DSet(a,b));
        PrintSet(stdout,"b - a",DSet(b,a));
        PrintSet(stdout,"a - c",DSet(a,c));
        PrintSet(stdout,"c - a",DSet(c,a));
        PrintSet(stdout,"a - e",DSet(a,e));
        PrintSet(stdout,"e - a",DSet(e,a));
        printf("memory used = %d\n",next-arena);
        PrintSet(stdout,"d U d",USet(d,d));
        PrintSet(stdout,"d U b",USet(d,b));
        PrintSet(stdout,"b U d",USet(b,d));
        PrintSet(stdout,"d ^ d",ISet(d,d));
        PrintSet(stdout,"d ^ b",ISet(d,b));
        PrintSet(stdout,"b ^ d",ISet(b,d));
        PrintSet(stdout,"d - d",DSet(d,d));
        PrintSet(stdout,"d - b",DSet(d,b));
        PrintSet(stdout,"b - d",DSet(b,d));
        printf("memory used = %d\n",next-arena);
        PrintSet(stdout,"f U g",USet(f,g));
        printf("memory used = %d\n",next-arena);
        PrintSet(stdout,"f ^ g",ISet(f,g));
        printf("memory used = %d\n",next-arena);
        PrintSet(stdout,"f - g",DSet(f,g));
        PrintSet(stdout,"g - f",DSet(g,f));
        printf("memory used = %d\n",next-arena);
        exit(0);
}
#endif
