/*
 *	gcc -o ray ray.c vectors.c -lm
 */
#include <stdio.h>
#include <math.h>
#include "vector.h"
#include "objs.h"
#include "data.h"
#include "interval.h"

int _trace = 0;

void
tron()
{
	fprintf(stderr,"\n<< Tracing begins >>\n");
	_trace=1;
}

void
troff()
{
	fprintf(stderr,"\n<< Tracing ends >>\n");
	_trace=0;
}

Set
intersectSph(l,s)
Ray l;
Sph s;
{
	Vector3d p;
	float	a,b,c,d,t1,t2,r;

	if (_trace) fprintf(stderr,
		"Entering intersectSph: l=Ray(s=(%g,%g,%g),d=(%g,%g,%g)), s=Sph(h=%d,c=(%g,%g,%g),r=%g)\n",
		l.s.x,l.s.y,l.s.z,l.d.x,l.d.y,l.d.z,s.hue,s.c.x,s.c.y,s.c.z,s.r);

	a = Dot3d(l.d,l.d);
	Diff3d(&p,l.s,s.c);
	b = 2 * Dot3d(l.d,p);
	c = Dot3d(p,p) - s.r*s.r;
	d = b*b - 4 * a * c;
	if (d >= epsilon) {
		d = sqrt(d);
		t1 = (-b + d) / (2*a);
		t2 = (-b - d) / (2*a);
		r = (t1 >= epsilon)
			? (t2 >= epsilon)
				? (t1<t2) ? t1 : t2
				: t1
			: (t2 >= epsilon) ? t2 : -1.0;
		if (_trace) fprintf(stderr, "Leaving intersectSph: %g\n", r);
		return MakeSet((t1<t2) ? t1 : t2,(t1>t2) ? t1 : t2);
	}
	if (_trace) fprintf(stderr, "Leaving intersectSph: missed\n");
	return NULL;
}

float
intersectPln(l,p)
Ray l;
Pln  p;
{
	register float	denom, t;
	if (_trace) fprintf(stderr,
		"Entering intersectPln: l=Ray(s=(%g,%g,%g),d=(%g,%g,%g)), p=Pln(h=%d,n=(%g,%g,%g),d=%g)\n",
		l.s.x,l.s.y,l.s.z,l.d.x,l.d.y,l.d.z,p.hue,p.n.x,p.n.y,p.n.z,p.d);
	denom = Dot3d(p.n,l.d);
	t = (abs(denom)>epsilon) ? -(p.d+Dot3d(p.n,l.s))/denom : -1;
	if (_trace) fprintf(stderr, "Leaving intersectPln: %g\n", t);
	return t;
}

void
RayTrace(Gaze, hue, intensity)
Ray Gaze;
int *hue;
float *intensity;
{
	Ray	Reflect,Shadow;
	Vector3d	v,normal,light;
	int	i0,i1,j;
	float	d,t0,t1;
	int	Illuminated;
	Set t;

	if (_trace) fprintf(stderr,
		"Entering RayTrace: Gaze=Ray(s=(%g,%g,%g),d=(%g,%g,%g))\n",
		Gaze.s.x,Gaze.s.y,Gaze.s.z,Gaze.d.x,Gaze.d.y,Gaze.d.z);

	i1 = -1;
	t1 = infinity;
	for (i0=0;i0<NbrSph;i0++)	/* find the closest object */
	{
		t = intersectSph(Gaze,Sphere[i0]);
		if (t) {
		t0 = t->start.val;
		if ((t0 > epsilon) && (t0 < t1))
		{
			i1 = i0;
			t1 = t0;
		}
		}
	}
	if (i1 >= 0) {			/* eye-ray hit something */
		SAdd3d(&v,t1,Gaze.d,Gaze.s);
		if (_trace) fprintf(stderr, "v=(%g,%g,%g))\n",v.x,v.y,v.z);
		Diff3d(&normal,v,Sphere[i1].c);
		Normalize3d(&normal);
		if (_trace) fprintf(stderr, "normal=(%g,%g,%g))\n",
				normal.x,normal.y,normal.z);
		if (Sphere[i1].hue == reflective) {  /* reflect */
			Reflect.s = v;
			d = -2*Dot3d(normal,Gaze.d);
			SAdd3d(&Reflect.d,d,normal,Gaze.d);
			/* recurse */
			*intensity = 0;
			RayTrace(Reflect,hue,intensity);
		} else {  /* opaque */
			*hue = Sphere[i1].hue;
			*intensity = ambient;
			Diff3d(&light,Source[0].p,v);
			Normalize3d(&light);
			Shadow.s = v; Shadow.d = light;
			Illuminated = 1;
			for (j=0;j<NbrSph;j++)
				if (i1 != j)
					if (intersectSph(Shadow,Sphere[j]))
						Illuminated = 0;
			d = Dot3d(light,normal);
			*intensity += ((d > 0) && Illuminated) ? d : 0;
		}
	} else {			  /* eye-ray missed the spheres */
		for (i0=0;i0<NbrPln;i0++)	/* find the closest object */
		{
			t0 = intersectPln(Gaze,Plane[i0]);
			if ((t0 > epsilon) && (t0 < t1)) {
				i1 = i0;
				t1 = t0;
			}
		}
		if (t1 == infinity) {	/* missed everything */
			*intensity = 0;
			*hue = 0;
			if (_trace) fprintf(stderr, "Leaving RayTrace...missed!\n");
			return;
		}
		SAdd3d(&v,t1,Gaze.d,Gaze.s);
		/* procedural definition of checkerboard pattern on walls */
		*hue = odd(round(v.x/200) + round(v.y/200) + round(v.z/200)) ? 1 : 2;
		/* procedural definition of ceiling and floor */
		if (v.y <-499) *hue = 5;
		if (v.y > 499) *hue = 6;

		*intensity = ambient;
		Diff3d(&light,Source[0].p,v);
		Normalize3d(&light);
		Shadow.s = v; Shadow.d = light;
		Illuminated = 1;
		for (j=0;j<NbrSph;j++)
			if (intersectSph(Shadow,Sphere[j]))
				Illuminated = 0;
		d = Dot3d(light,Plane[i1].n);
		*intensity += ((d > 0) && Illuminated) ? d : 0;
	}
	if (_trace) fprintf(stderr,"Leaving RayTrace: hue=%d, intensity=%g)\n",
		*hue,*intensity);
}

void
BuildImage (width, height, xstart, ystart, incr)
int width, height;
{
	int	row, col, hue;
	float	intensity;
	Ray	Gaze;
	float sc = 2.0 / (width>height?width:height);
	float tx = sc * width / 2;
	float ty = sc * height / 2;

	Make3d(&(Gaze.s),0.0,0.0,-200.0);	 /* source of gaze */
	for (col=xstart;col<width;col+=incr) {
		for (row=ystart;row<height;row+=incr) {
			/* direction of gaze */
			Make3d(&(Gaze.d),sc*col-tx,sc*row-ty,1.0);
			intensity = 0;
reset_sets();
			RayTrace(Gaze,&hue,&intensity);
			if (intensity<0.0) intensity = 0.0;
			if (intensity>1.0) intensity = 1.0;
			intensity=1-(1-intensity)*(1-intensity);
			Dither(col,row,hue,intensity);
		}
	}
}

void
BuildPixel (width, height, col, row)
int width, height, col, row;
{
	int	hue;
	float	intensity;
	Ray	Gaze;

	Make3d(&Gaze.s,0.0,0.0,-height);	 /* source of gaze */
	/* direction of gaze */
	Make3d(&Gaze.d,(float)(col-width/2),(float)(row-height/2),height);
	RayTrace(Gaze,&hue,&intensity);
	if (intensity<0.0) intensity = 0.0;
	if (intensity>1.0) intensity = 1.0;
	intensity=1-(1-intensity)*(1-intensity);
	Dither(col,row,hue,intensity);
}
