/*vector.h*/
#include <math.h>

typedef struct {float x,y,z;int w;} Vector3d;

#ifdef LINT
void  Make3d(Vector3d *r, float xx, float yy, float zz);
void  Negate3d(Vector3d *r);
void  Scale3d(Vector3d *r, float A, Vector3d b);
void  Sum3d(Vector3d *r, Vector3d a);
void  SSum3d(Vector3d *r, float A, Vector3d a);
void  Add3d(Vector3d *r, Vector3d a, Vector3d b);
void  SAdd3d(Vector3d *r, float A, Vector3d a, Vector3d b);
void  SSAdd3d(Vector3d *r, float A, Vector3d a, float B, Vector3d b);
void  Diff3d(Vector3d *r, Vector3d a, Vector3d b);
void  Cross3d(Vector3d *r, Vector3d a, Vector3d b);
float Dot3d(Vector3d a, Vector3d b);
float Length3d(Vector3d v);
void  Normalize3d(Vector3d *r);
#else
#define Make3d(r,xx,yy,zz) ((r)->x=(xx),(r)->y=(yy),(r)->z=(zz))
#define Negate3d(r) ((r)->x=-(r)->x,(r)->y=-(r)->y,(r)->z=-(r)->z)
#define Scale3d(r,A,b) ((r)->x=A*(b).x,(r)->y=A*(b).y,(r)->z=A*(b).z)
#define Sum3d(r,a) ((r)->x+=(a).x,(r)->y+=(a).y,(r)->z+=(a).z)
#define SSum3d(r,A,a) ((r)->x+=(A)*(a).x,(r)->y+=(A)*(a).y,(r)->z+=(A)*(a).z)
#define Add3d(r,a,b) ((r)->x=(a).x+(b).x,(r)->y=(a).y+(b).y,(r)->z=(a).z+(b).z)
#define SAdd3d(r,A,a,b) ((r)->x=(A)*(a).x+(b).x,\
					     (r)->y=(A)*(a).y+(b).y,\
					     (r)->z=(A)*(a).z+(b).z)
#define SSAdd3d(r,A,a,B,b) ((r)->x=(A)*(a).x+(B)*(b).x,\
					        (r)->y=(A)*(a).y+(B)*(b).y,\
					        (r)->z=(A)*(a).z+(B)*(b).z)
#define Diff3d(r,a,b) ((r)->x=(a).x-(b).x,(r)->y=(a).y-(b).y,(r)->z=(a).z-(b).z)
#define Cross3d(r,a,b) ((r)->x=(a).y*(b).z-(a).z*(b).y,\
					    (r)->y=(a).z*(b).x-(a).x*(b).z,\
					    (r)->z=(a).x*(b).y-(a).y*(b).x)
#define Dot3d(a,b) ((a).x*(b).x+(a).y*(b).y+(a).z*(b).z)
#define Length3d(v) sqrt(Dot3d(v,v))
#define Normalize3d(r) {float l=Dot3d(*r,*r);if(l!=1.0)Scale3d(r,1/sqrt(l),*r);}
#endif

typedef struct {float u,v;int w;} Vector2d;

#ifdef LINT
void  Make2d(Vector2d *r, float uu, float vv);
void  Negate2d(Vector2d *r);
void  Scale2d(Vector2d *r, float A, Vector2d b);
void  Sum2d(Vector2d *r, Vector2d a);
void  SSum2d(Vector2d *r, float A, Vector2d a);
void  Add2d(Vector2d *r, Vector2d a, Vector2d b);
void  SAdd2d(Vector2d *r, float A, Vector2d a, Vector2d b);
void  SSAdd2d(Vector2d *r, float A, Vector2d a, float B, Vector2d b);
void  Diff2d(Vector2d *r, Vector2d a, Vector2d b);
float Dot2d(Vector2d a, Vector2d b);
float Length2d(Vector2d v);
void  Normalize2d(Vector2d *r);
#else
#define Make2d(r,uu,vv) ((r)->u=(uu),(r)->v=(vv))
#define Negate2d(r) ((r)->u=-(r)->u,(r)->v=-(r)->v)
#define Scale2d(r,A,b) ((r)->u=A*(b).u,(r)->v=A*(b).v)
#define Sum2d(r,a) ((r)->u+=(a).u,(r)->v+=(a).v)
#define SSum2d(r,A,a) ((r)->u+=(A)*(a).u,(r)->v+=(A)*(a).v)
#define Add2d(r,a,b) ((r)->u=(a).u+(b).u,(r)->v=(a).v+(b).v)
#define SAdd2d(r,A,a,b) ((r)->u=(A)*(a).u+(b).u,\
					     (r)->v=(A)*(a).v+(b).v)
#define SSAdd2d(r,A,a,B,b) ((r)->u=(A)*(a).u+(B)*(b).u,\
					        (r)->v=(A)*(a).v+(B)*(b).v)
#define Diff2d(r,a,b) ((r)->u=(a).u-(b).u,(r)->v=(a).v-(b).v)
#define Dot2d(a,b) ((a).u*(b).u+(a).v*(b).v)
#define Length2d(v) sqrt(Dot2d(v,v))
#define Normalize2d(r) {float l=Dot2d(*r,*r);if(l!=1.0)Scale2d(r,1/sqrt(l),*r);}
#endif
